import os

from amino.map import Map


class EnvVars(object):

    @property
    def vars(self):
        return Map(os.environ)

    def __contains__(self, name):
        return name in self.vars

    def __getitem__(self, name):
        return self.vars.get(name)

    def __setitem__(self, name, value):
        self.vars[name] = value

env = EnvVars()

__all__ = ('EnvVars', 'env')



from typing import TypeVar, Generic, Callable, Union, Any
from typing import Tuple  # NOQA
from functools import wraps, partial
from operator import eq, is_not
import inspect
import traceback

from amino import boolean
from amino.tc.base import Implicits
from amino.func import call_by_name, I

A = TypeVar('A')
B = TypeVar('B')


class Maybe(Generic[A], Implicits, implicits=True):

    __slots__ = ()

    def __new__(tp, value: A, checker=partial(is_not, None)):
        return Maybe.check(value, checker)

    @staticmethod
    def check(value: A, checker=partial(is_not, None)):
        return Just(value) if checker(value) else Empty()

    @staticmethod
    def from_call(f: Callable[..., A], *args, **kwargs):
        exc = kwargs.pop('exc', Exception)
        try:
            return Maybe.check(f(*args, **kwargs))
        except exc:
            if exc == Exception:
                from amino.logging import log
                frame = inspect.currentframe().f_back  # type: ignore
                stack = traceback.format_stack(frame)
                log.exception('Maybe.from_call:')
                log.error(''.join(stack))
            return Empty()

    @staticmethod
    def typed(value: A, tpe: type):
        return Maybe.check(value, lambda a: isinstance(a, tpe))

    @staticmethod
    def wrap(mb: Union['Maybe[A]', None]):
        return mb if mb is not None and isinstance(mb, Just) else Empty()

    @property
    def _get(self) -> Union[A, None]:
        pass

    def cata(self, f: Callable[[A], B], b: Union[B, Callable[[], B]]) -> B:
        return (
            f(self._get)  # type: ignore
            if self.is_just
            else call_by_name(b)
        )

    def filter(self, f: Callable[[A], B]):
        l = lambda a: self if f(a) else Empty()
        return self.flat_map(l)

    def get_or_else(self, a: Union[A, Callable[[], A]]):
        return self.cata(I, a)

    __or__ = get_or_else

    def get_or_raise(self, e: Exception):
        def raise_e():
            raise e
        return self.cata(I, raise_e)

    def get_or_fail(self, err):
        return self.get_or_raise(Exception(call_by_name(err)))

    def __contains__(self, v):
        return self.contains(v)

    def foreach(self, f: Callable[[A], Any]):
        self.cata(f, None)

    def error(self, f: Callable[[], Any]) -> 'Maybe[A]':
        self.cata(I, f)
        return self

    def observe(self, f: Callable[[A], Any]):
        self.foreach(f)
        return self

    effect = observe

    def __iter__(self):
        return iter(self.to_list)

    @property
    def is_just(self):
        return boolean.Boolean(isinstance(self, Just))

    @property
    def is_empty(self):
        return not self.is_just

    empty = is_empty

    def __nonzero__(self):
        return self.is_just

    @property
    def to_list(self):
        from amino.list import List
        return self.cata(lambda v: List(v), List())

    @property
    async def unsafe_await(self):
        if self.is_just:
            ret = await self._get()
            return Maybe(ret)
        else:
            return self

    async def unsafe_await_or(self, b: Union[B, Callable[[], B]]):
        return (Maybe(await(self._get)) if self.is_just  # type: ignore
                else call_by_name(b))

    @property
    def contains_coro(self):
        return self.exists(inspect.iscoroutine)

    @property
    def json(self):
        return self.cata(I, lambda: None)


class Just(Maybe):

    __slots__ = 'x',

    def __new__(tp, value: A, *args, **kwargs):
        return object.__new__(tp)

    def __init__(self, value):
        self.x = value

    @property
    def _get(self) -> Union[A, None]:
        return self.x

    def __str__(self):
        return 'Just({!s})'.format(self.x)

    def __repr__(self):
        return 'Just({!r})'.format(self.x)

    def __eq__(self, other):
        if not isinstance(other, Just):
            return False
        return eq(self.x, other.x)

    def __hash__(self):
        return hash(self._get)


class Empty(Maybe):

    __object = None  # type: Empty

    def __new__(tp, *args, **kwargs):
        if Empty.__object is None:
            Empty.__object = object.__new__(tp)
        return Empty.__object

    def __str__(self):
        return 'Empty()'

    __repr__ = __str__

    def __eq__(self, other):
        return isinstance(other, Empty)

    def __hash__(self):
        return hash('Empty')


def may(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return Maybe.check(f(*args, **kwargs))
    return wrapper


def flat_may(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        res = f(*args, **kwargs)
        return res if isinstance(res, Maybe) else Maybe(res)
    return wrapper

__all__ = ('Maybe', 'Just', 'Empty', 'may')



from functools import wraps, partial
from inspect import getfullargspec
from typing import Callable, Union, Any

import fn


class F(fn.F):

    def __truediv__(self, f):
        return self >> (lambda a: a / f)

    def __floordiv__(self, f):
        return self >> (lambda a: a // f)

    @property
    def name(self):
        from amino.anon import AnonCallable
        f = (
            self.f.func
            if isinstance(self.f.func, AnonCallable) else
            self.f.func.__name__
        ) if self._is_partial else self.f.__name__
        return str(f)

    @property
    def _is_partial(self):
        return isinstance(self.f, partial)

    def __str__(self):
        from amino.anon import format_funcall
        rep = (format_funcall(self.name, self.f.args, self.f.keywords)
               if self._is_partial
               else '{}()'.format(self.name))
        return 'F({})'.format(rep)


def curried(func):
    @wraps(func)
    def _curried(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func
        spec = getfullargspec(f)
        if count == len(spec.args) - len(args):
            return func(*args, **kwargs)
        else:
            return curried(partial(func, *args, **kwargs))
    return _curried


class Identity:

    def __call__(self, a):
        return a

    def __str__(self):
        return '(a => a)'

I = Identity()


class Val:

    def __init__(self, value) -> None:
        self.value = value

    def __call__(self):
        return self.value

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.value)


def flip(a, b):
    return b, a

CallByName = Union[Any, Callable[[], Any]]


def call_by_name(b: CallByName):
    return b() if callable(b) else b  # type: ignore

__all__ = ('curried', 'F', 'I', 'flip', 'call_by_name', 'Val')



from pathlib import Path

import sure

from amino import Maybe, Empty, Just


class AssBuilder(sure.AssertionBuilder):

    @sure.assertionmethod
    def contain(self, what):
        if isinstance(self.obj, Maybe):
            return self.just_contain(what)
        else:
            return super().contain(what)

    @sure.assertionmethod
    def just_contain(self, what):
        self.obj.should.be.a(Maybe)
        if self.negative:
            msg = "{} contains {}, expected {}"
            assert self.obj.is_empty or self.obj._get != what,\
                msg.format(self.obj, self.obj._get, what)
        else:
            self.be.just
            self.obj._get.should.equal(what)
        return True

    def _bool(self, pred, agent, action):
        no = ('not ' if self.negative else '') + 'to '
        assert pred(self.obj) ^ self.negative,\
            'expected {} \'{}\' {} {}'.format(agent, self.obj, no, action)
        return True

    @sure.assertionproperty
    def just(self):
        return self.be.a(Just)

    @sure.assertionproperty
    def empty(self):
        if isinstance(self.obj, Maybe):
            return self.be.a(Empty)
        else:
            return super().empty

    @sure.assertionproperty
    def exist(self):
        err = "can only check existence of Path, not {}"
        assert isinstance(self.obj, Path), err.format(self.obj)
        return self._bool(lambda a: a.exists(), "path", "exist")

    @sure.assertionmethod
    def start_with(self, prefix):
        return self.match('^{}'.format(prefix))


def install_assertion_builder(builder):
    sure.AssertionBuilder = builder

__all__ = ('install_assertion_builder', 'AssBuilder')



from typing import TypeVar, Dict, Generic, Tuple, Callable

from toolz import dicttoolz

from amino import Maybe, may, Just
from amino.list import List
from amino.boolean import Boolean

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')


class Map(Dict[A, B], Generic[A, B]):  # type: ignore

    @staticmethod
    def wrap(d: Dict[A, B]) -> 'Map[A, B]':
        return Map(d)

    @may
    def get(self, key):
        return Dict.get(self, key)

    def get_item(self, key):
        return self.get(key) / (lambda a: (key, a))

    def get_all(self, *keys):
        def append(zm, k):
            return zm // (lambda z: self.get(k) / z.cat)
        return List.wrap(keys).fold_left(Just(List()))(append)

    def get_all_map(self, *keys):
        def append(zm, k):
            return zm // (lambda z: self.get_item(k) / z.cat)
        return List.wrap(keys).fold_left(Just(Map()))(append)

    def get_or_else(self, key, default: Callable[[], C]):
        return self.get(key).get_or_else(default)

    def set_if_missing(self, key: A, default: Callable[[], B]):
        if key in self:
            return self
        else:
            return self + (key, default())

    def __str__(self):
        return str(dict(self))

    def cat(self, item: Tuple[A, B]):
        return Map(dicttoolz.assoc(self, *item))

    __add__ = cat

    def merge(self, other: 'Map[A, B]'):
        return Map(dicttoolz.merge(self, other))

    __pow__ = merge

    def __sub__(self, key: A):
        return Map(dicttoolz.dissoc(self, key))

    def find(self, f: Callable[[B], bool]) -> Maybe[Tuple[A, B]]:
        return self.to_list.find(lambda item: f(item[1]))

    def find_key(self, f: Callable[[A], bool]) -> Maybe[Tuple[A, B]]:
        return self.to_list.find(lambda item: f(item[0]))

    def valfilter(self, f: Callable[[B], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.valfilter(f, self))

    def keyfilter(self, f: Callable[[A], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.keyfilter(f, self))

    def filter(self, f: Callable[[Tuple[A, B]], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.itemfilter(f, self))

    def valmap(self, f: Callable[[B], C]) -> 'Map[A, C]':
        return Map(dicttoolz.valmap(f, dict(self)))

    def keymap(self, f: Callable[[A], C]) -> 'Map[C, B]':
        return Map(dicttoolz.keymap(f, dict(self)))

    def map(self, f: Callable[[A, B], Tuple[C, D]]) -> 'Map[C, D]':
        wrap = lambda a: f(*a)
        return Map(dicttoolz.itemmap(wrap, self))

    def flat_map(
            self,
            f: Callable[[A, B], Maybe[Tuple[C, D]]]
    ) -> 'Map[C, D]':
        filtered = List.wrap([f(a, b) for a, b in self.items()])\
            .join
        return Map(filtered)

    def bimap(self, fa: Callable[[A], C], fb: Callable[[B], D]) -> 'Map[C, D]':
        return self.map(lambda a, b: (fa(a), fb(b)))

    def map2(self, f: Callable[[A, B], C]) -> List[C]:
        return List.wrap([f(a, b) for a, b in self.items()])

    @property
    def to_list(self):
        return List.wrap(self.items())

    @property
    def head(self):
        return self.to_list.head

    @property
    def keys_view(self):
        return Dict.keys(self)

    @property
    def values_view(self):
        return Dict.values(self)

    @property
    def k(self):
        return List(*Dict.keys(self))

    @property
    def v(self):
        return List(*Dict.values(self))

    @property
    def is_empty(self):
        return self.k.is_empty

    empty = is_empty

    def at(self, *keys):
        return self.keyfilter(lambda a: a in keys)

    def values_at(self, *keys):
        return List.wrap(keys) // self.get

    def has_key(self, name):
        return Boolean(name in self)

    contains = has_key

    def add_maybe(self, key: A, value: Maybe[B]):
        return value / (lambda a: self + (key, a)) | self

__all__ = ('Map',)



import os
import time
import shutil
import inspect
import warnings
import traceback
from datetime import datetime
from functools import wraps
from contextlib import contextmanager

import spec

import amino
from amino.logging import amino_stdout_logging, Logging
from amino.test.sure_ext import install_assertion_builder, AssBuilder
from amino.test import path
from amino import Path, List


default_timeout = 20 if 'TRAVIS' in os.environ else 3


def later(ass, *a, timeout=None, intval=0.1, **kw):
    timeout = default_timeout if timeout is None else timeout
    start = datetime.now()
    ok = False
    while not ok and (datetime.now() - start).total_seconds() < timeout:
        try:
            ass(*a, **kw)
            ok = True
        except AssertionError:
            time.sleep(intval)
    return ass(*a, **kw)


class Spec(spec.Spec, Logging):

    def __init__(self):
        self._warnings = True

    def setup(self):
        if path.__base_dir__:
            shutil.rmtree(str(path.temp_path()), ignore_errors=True)
        if self._warnings:
            warnings.resetwarnings()
        amino.development = True
        amino_stdout_logging()
        install_assertion_builder(AssBuilder)

    def teardown(self, *a, **kw):
        warnings.simplefilter('ignore')

    def _wait_for(self, pred, timeout=default_timeout, intval=0.1):
        start = datetime.now()
        while (not pred() and
               (datetime.now() - start).total_seconds() < timeout):
            time.sleep(intval)
        pred().should.be.ok

    def _wait(self, seconds):
        time.sleep(seconds)


class IntegrationSpec(Spec):

    def setup(self):
        os.environ['AMINO_INTEGRATION'] = '1'
        amino.integration_test = True
        super().setup()


def profiled(sort='time'):
    fname = 'prof'
    def dec(f):
        import cProfile
        import pstats
        @wraps(f)
        def wrap(*a, **kw):
            cProfile.runctx('f(*a, **kw)', dict(), dict(f=f, a=a, kw=kw),
                            filename=fname)
            stats = pstats.Stats(fname)
            stats.sort_stats(sort).print_stats(30)
            Path(fname).unlink()
        return wrap
    return dec


def callers(limit=20):
    stack = (List.wrap(inspect.stack())
             .filter_not(lambda a: 'amino' in a.filename))
    data = stack[:limit] / (lambda a: a[1:-2] + tuple(a[-2]))
    return ''.join(traceback.format_list(data))


def timed(f):
    @wraps(f)
    def wrap(*a, **kw):
        import time
        start = time.time()
        v = f(*a, **kw)
        from ribosome.logging import log
        log.info('{}: {}'.format(f.__name__, time.time() - start))
        return v
    return wrap


@contextmanager
def timer(name='timer'):
    import time
    start = time.time()
    v = yield
    from ribosome.logging import log
    log.info('{}: {}'.format(name, time.time() - start))
    return v

__all__ = ('Spec', 'IntegrationSpec', 'profiled', 'later', 'timed', 'timer')



from typing import TypeVar, Generic, Callable, Union
import functools

A = TypeVar('A')
_attr_fmt = '_{}__value'


class LazyMeta(type):

    def __new__(mcs, name, bases, dct):
        lazies = [k for k, v in dct.items() if isinstance(v, lazy)]
        attrs = tuple(map('_{}__value'.format, lazies))
        inst = super().__new__(mcs, name, bases, dct)
        inst.__slots__ = attrs
        return inst


class LazyError(Exception):

    def __init__(self, name, inst):
        self.name = name
        self.inst = inst

    def __str__(self):
        msg = (
            'class {} with lazy attribute must have a __dict__,  a {}' +
            ' slot or use LazyMeta as metaclass'
        )
        return msg.format(type(self.inst).__name__, self.name)


class lazy(Generic[A]):

    def __init__(self, func: Callable[..., A], name=None) -> None:
        self.func = func
        self._attr_name = _attr_fmt.format(name or self.func.__name__)
        functools.wraps(self.func)(self)  # type: ignore

    def __get__(self, inst, inst_cls) -> Union[A, 'lazy[A]']:
        return self if inst is None else self._get(inst, inst_cls)

    def _get(self, inst, inst_cls):
        self._check(inst)
        return getattr(inst, self._attr_name)

    def _valid_slot(self, inst):
        return hasattr(inst, '__slots__') and self._attr_name in inst.__slots__

    def _check(self, inst):
        if not (hasattr(inst, '__dict__') or self._valid_slot):
            self._complain()
        if not hasattr(inst, self._attr_name):
            object.__setattr__(inst, self._attr_name, self.func(inst))

    def _complain(self, inst):
        raise LazyError(self._attr_name, inst)


class Lazy(metaclass=LazyMeta):
    pass

__all__ = ('lazy',)



import abc
from typing import Generic, Callable, Any, TypeVar

from amino import L, _

A = TypeVar('A')


class Transformer(Generic[A], metaclass=abc.ABCMeta):

    def __init__(self, val: A) -> None:
        self.val = val

    @abc.abstractmethod
    def pure(self, b) -> A:
        ...

    def flat_map(self, f: Callable[[A], Any]) -> 'Transformer':
        return self.__class__(f(self.val))  # type: ignore

    __floordiv__ = flat_map

    def map(self, f: Callable[[A], Any]):
        return self.flat_map(L(f)(_) >> self.pure)

    __truediv__ = map

    def effect(self, f: Callable[[A], Any]):
        f(self.val)
        return self

    __mod__ = effect

    def effect0(self, f: Callable[[], Any]):
        f()
        return self

    __matmul__ = effect0

__all__ = ('Transformer',)



import random
import string


class Random(object):
    _char_pool = string.ascii_uppercase + string.digits

    @staticmethod
    def string(n=10):
        return ''.join(random.choice(Random._char_pool) for _ in range(n))

__all__ = ('Random',)



from typing import TypeVar, Callable, Tuple

from amino.tc.base import tc_prop, ImplicitInstances
from amino.tc.optional import Optional
from amino.tc.monad import Monad
from amino.tc.traverse import Traverse
from amino.lazy import lazy
from amino.maybe import Just, Empty
from amino.either import Right, Either, Left
from amino.tc.applicative import Applicative
from amino.tc.foldable import Foldable
from amino import curried

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class EitherInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino.map import Map
        return Map(
            {
                Monad: EitherMonad(),
                Optional: EitherOptional(),
                Traverse: EitherTraverse(),
                Foldable: EitherFoldable(),
            }
        )


class EitherMonad(Monad):

    def pure(self, a: A):
        return Right(a)

    def flat_map(self, fa: Either[A, B], f: Callable[[B], Either[A, C]]
                 ) -> Either[A, C]:
        return f(fa.value) if isinstance(fa, Right) else fa


class EitherOptional(Optional):

    @tc_prop
    def to_maybe(self, fa: Either):
        return Just(fa.value) if fa.is_right else Empty()

    def to_either(self, fa: Either, left):
        return fa

    @tc_prop
    def present(self, fa: Either):
        return fa.is_right


class EitherTraverse(Traverse):

    def traverse(self, fa: Either[A, B], f: Callable, tpe: type):
        monad = Applicative[tpe]
        r = lambda a: monad.map(f(a), Right)
        return fa.cata(lambda a: monad.pure(Left(a)), r)


class EitherFoldable(Foldable):

    @tc_prop
    def with_index(self, fa: Either[A, B]) -> Either[A, Tuple[int, B]]:
        return Right(0) & fa

    def filter(self, fa: Either[A, B], f: Callable[[B], bool]):
        return fa // (lambda a: Right(a) if f(a) else Left('filtered'))

    @curried
    def fold_left(self, fa: Either[A, B], z: C, f: Callable[[C, B], C]) -> C:
        return fa / (lambda a: f(z, a)) | z

    def find(self, fa: Either[A, B], f: Callable[[B], bool]):
        return fa.to_maybe.find(f)

    def find_map(self, fa: Either[A, B], f: Callable[[B], Either[A, C]]
                 ) -> Either[A, C]:
        return fa // f

    def index_where(self, fa: Either[A, B], f: Callable[[B], bool]):
        return fa.to_maybe.index_where(f)

__all__ = ('EitherInstances',)



import abc
from typing import TypeVar, Generic, Callable
import operator

from lenses import lens, Lens

from amino.tc.base import TypeClass
from amino.tc.functor import Functor
from amino.func import curried
from amino.maybe import Maybe, Empty, Just
from amino.boolean import Boolean
from amino import _

F = TypeVar('F')
G = TypeVar('G')
A = TypeVar('A')
B = TypeVar('B')
Z = TypeVar('Z')


class Foldable(Generic[F], TypeClass):
    # FIXME lens functions return index lenses, which is not a property of
    # Foldable

    @abc.abstractmethod
    def with_index(self, fa: F) -> F:
        ...

    @abc.abstractmethod
    def filter(self, fa: F, f: Callable[[A], bool]) -> F:
        ...

    def filter_not(self, fa: F, f: Callable[[A], bool]) -> F:
        pred = lambda a: not f(a)
        return self.filter(fa, pred)

    def filter_type(self, fa: F, tpe: type) -> F:
        pred = lambda a: isinstance(a, tpe)
        return self.filter(fa, pred)

    @abc.abstractmethod
    def find(self, fa: F, f: Callable[[A], bool]) -> Maybe[A]:
        ...

    @abc.abstractmethod
    @curried
    def fold_left(self, fa: F, z: Z, f: Callable[[Z, A], Z]) -> Z:
        ...

    def fold_map(self, fa: F, z: B, f: Callable[[A], B],
                 g: Callable[[Z, B], Z]=operator.add) -> Z:
        ''' map `f` over the traversable, then fold over the result
        using the supplied initial element `z` and operation `g`,
        defaulting to addition for the latter.
        '''
        mapped = Functor[type(fa)].map(fa, f)
        return self.fold_left(mapped)(z)(g)

    @abc.abstractmethod
    def find_map(self, fa: F, f: Callable[[A], Maybe[B]]) -> Maybe[B]:
        ...

    def find_type(self, fa: F, tpe: type):
        pred = lambda a: isinstance(a, tpe)
        return self.find(fa, pred)

    @abc.abstractmethod
    def index_where(self, fa: F, f: Callable[[A], bool]) -> Maybe[int]:
        ...

    def index_of(self, fa: F, a: A) -> Maybe[int]:
        return self.index_where(fa, _ == a)

    def exists(self, fa: F, f: Callable[[A], bool]) -> Boolean:
        return Boolean(self.find(fa, f).is_just)

    def contains(self, fa, v):
        return self.exists(fa, _ == v)

    def lens(self, fa: F, f: Callable[[A], bool]) -> Maybe[Lens]:
        return self.index_where(fa, f) / (lambda i: lens()[i])

    def find_lens(self, fa: F, f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
        check = lambda a: f(a[1]) / (lambda b: (a[0], b))
        index = lambda i, l: lens()[i].add_lens(l)
        wi = self.with_index(fa)
        return self.find_map(wi, check).map2(index)

    def find_lens_pred(self, fa: F, f: Callable[[A], bool]) -> Maybe[Lens]:
        g = lambda a: Boolean(f(a)).maybe(lens())
        return self.find_lens(fa, g)

    def _min_max(self, fa: F, f: Callable[[A], int],
                 pred: Callable[[A, A], A]) -> Maybe[A]:
        def folder(z, a):
            return (
                z.map(lambda b: b if pred(f(b), f(a)) else a)
                .or_else(Just(a))
            )
        return self.fold_left(fa, Empty())(folder)

    def max_by(self, fa: F, f: Callable[[A], int]) -> Maybe[A]:
        return self._min_max(fa, f, operator.gt)

    def min_by(self, fa: F, f: Callable[[A], int]) -> Maybe[A]:
        return self._min_max(fa, f, operator.lt)

__all__ = ('Foldable',)



import abc
from typing import TypeVar, Generic, Callable

from amino.tc.base import TypeClass
from amino.func import I

F = TypeVar('F')
G = TypeVar('G')
A = TypeVar('A')
B = TypeVar('B')
Z = TypeVar('Z')


class Traverse(Generic[F], TypeClass):
    # FIXME lens functions return index lenses, which is not a property of
    # Traverse

    @abc.abstractmethod
    def traverse(self, fa: F, f: Callable, tpe: type):
        ...

    def sequence(self, fa: F, tpe: type):
        return self.traverse(fa, I, tpe)

__all__ = ('Traverse',)



import pathlib

from amino.maybe import Maybe, Just, Empty, may, flat_may
from amino.list import List
from amino.lazy_list import LazyList
from amino.map import Map
from amino.either import Left, Right, Either
from amino.future import Future
from amino.boolean import Boolean
from amino.func import curried, F, I
from amino.anon import __, L, _
from amino.env_vars import env
from amino.task import Try, Task
from amino.logging import Logger, log
from amino.eff import Eff

development = 'AMINO_DEVELOPMENT' in env
integration_test = 'AMINO_INTEGRATION' in env

Path = pathlib.Path

__all__ = ('Maybe', 'Just', 'Empty', 'may', 'List', 'Map', '_', 'Future',
           'Boolean', 'development', 'flat_may', 'curried', '__', 'F', 'Left',
           'Right', 'Either', 'env', 'Try', 'LazyList', 'Logger', 'log', 'I',
           'L', 'Eff', 'Task')



from typing import TypeVar, Generic, Callable

from amino.tc.base import TypeClass
from amino.tc.monad import Monad

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Tap(Generic[F], TypeClass):

    def tap(self, fa: F, f: Callable[[A], B]) -> F:
        def effect(a):
            f(a)
            return fa
        return Monad[type(fa)].flat_map(fa, effect)  # type: ignore

    def __mod__(self, fa, f):
        return self.tap(fa, f)

__all__ = ('Tap',)



import importlib
import abc
from typing import Dict, Callable
from typing import GenericMeta  # type: ignore
from functools import partial, wraps

from amino.util.string import snake_case
from amino.lazy import lazy
from amino.tc.show import Show

from amino.logging import Logging


class TypeClassMeta(GenericMeta):

    def __getitem__(self, tpe: type):
        return Instances.lookup(self, tpe)

    def exists(self, tpe: type):
        try:
            self[tpe]
        except ImplicitNotFound:
            return False
        else:
            return True


class TypeClass(object, metaclass=TypeClassMeta):
    pass


class TypeClasses(object):

    @property
    def instances(self):
        pass


class ImplicitInstancesNotFound(Exception):

    def __init__(self, mod, cls, name):
        msg = 'invalid implicit class path {}.{} for {}'.format(mod, cls, name)
        super().__init__(msg)


operators = (
    '__floordiv__',
    '__truediv__',
    '__mod__',
    '__or__',
    '__and__',
)


class ImplicitsMeta(GenericMeta):

    def _infer_implicits(name):
        snake = snake_case(name)
        return 'amino.instances.{}'.format(snake), '{}Instances'.format(name)

    def _mk_operator(name):
        def dispatch(self, other):
            return self._operator(name, other)
        return dispatch

    def _attach_operators(inst):
        for op in operators:
            setattr(inst, op, ImplicitsMeta._mk_operator(op))

    def __new__(cls, name, bases, namespace, imp_mod=None, imp_cls=None,
                implicits=False, **kw):
        inst = super().__new__(cls, name, bases, namespace, **kw)
        if not implicits:
            return inst
        else:
            if imp_mod is None or imp_cls is None:
                imp_mod, imp_cls = ImplicitsMeta._infer_implicits(name)
            inst.imp_mod = imp_mod
            inst.imp_cls = imp_cls
            inst.name = name
            inst._implicits_instance = None
            ImplicitsMeta._attach_operators(inst)
            Instances.add(name, inst)
            return inst

    @property
    def _implicits(self):
        m = self.imp_mod
        c = self.imp_cls
        err = ImplicitInstancesNotFound(m, c, self.name)
        try:
            mod = importlib.import_module(m)
        except ImportError:
            raise err
        else:
            if hasattr(mod, c):
                instances = getattr(mod, c)()
                return instances
            else:
                raise err

    @property
    def implicits(self):
        if self._implicits_instance is None:
            self._implicits_instance = self._implicits.instances
        return self._implicits_instance


def tc_prop(f):
    f._tc_prop = True
    return f


class Implicits(Logging, metaclass=ImplicitsMeta):
    permanent = True

    def _lookup_implicit_attr(self, name):
        return next((getattr(inst, name)
                     for inst in type(self).implicits.v
                     if hasattr(inst, name)), None)

    def _bound_implicit_attr(self, name):
        f = self._lookup_implicit_attr(name)
        if f is not None:
            if hasattr(f, '_tc_prop'):
                return f(self)
            else:
                return partial(f, self)

    def _set_implicit_attr(self, name):
        f = self._lookup_implicit_attr(name)
        if f is not None:
            @wraps(f)
            def wrap(self, *a, **kw):
                return f(self, *a, **kw)
            g = property(wrap) if hasattr(f, '_tc_prop') else wrap
            setattr(type(self), name, g)
            return getattr(self, name)

    def __getattr__(self, name):
        imp = (self._set_implicit_attr(name) if Implicits.permanent else
               self._bound_implicit_attr(name))
        if imp is None:
            err = '\'{}\' has no attribute \'{}\''.format(self, name)
            raise AttributeError(err)
        else:
            return imp

    def _operator(self, name, other):
        op = (self._set_implicit_attr(name) if Implicits.permanent else
              self._bound_implicit_attr(name))
        if op is None:
            err = '\'{}\' has no implicit operator \'{}\''.format(self, name)
            raise TypeError(err)
        else:
            return op(other)

    @property
    def dbg(self):
        v = self.log.verbose
        v(self)
        return self

    @property
    def dbgr(self):
        v = self.log.verbose
        v(repr(self))
        return self


class ImplicitInstances(object, metaclass=abc.ABCMeta):

    @lazy
    def instances(self):
        return (TC.instances ** self._instances **
                self._override_instances)

    @abc.abstractproperty  # type: ignore
    def _instances(self) -> Dict[type, TypeClass]:
        ...

    @property
    def _override_instances(self) -> Dict[type, TypeClass]:
        from amino.map import Map
        return Map()


class GlobalTypeClasses(TypeClasses):

    @property
    def instances(self):
        from amino.map import Map
        from amino.tc.tap import Tap
        return Map({Show: Show(), Tap: Tap()})


TC = GlobalTypeClasses()


class ImplicitNotFound(Exception):

    def __init__(self, tc, a):
        msg = 'no type class found for {}[{}]'.format(tc, a)
        super().__init__(msg)


class AllInstances(object):

    def __init__(self):
        self._instances = dict()

    def add(self, name, inst: Callable[[], ImplicitInstances]):
        self._instances[name] = inst

    def lookup(self, TC, G):
        ''' Find an instance of the type class `TC` for type `G`.
        Iterates `G`'s parent classes, looking up instances for each,
        checking whether the instance is a subclass of the target type
        class @`C`.
        '''
        from amino.lazy_list import LazyList
        from amino.anon import _, L
        match = L(self._lookup_type)(TC, _)
        result = LazyList(map(match, G.__mro__))\
            .find(_.is_just)\
            .join\
            .get_or_raise(ImplicitNotFound(TC, G))
        return result[1]

    def _lookup_type(self, TC, G):
        from amino.maybe import Empty
        if G.__name__ in self._instances:
            match = lambda I: isinstance(I, TC)
            return self._instances[G.__name__].implicits.find(match)
        else:
            return Empty()

Instances = AllInstances()  # type: AllInstances

__all__ = ('TypeClasses', 'TC', 'tc_prop', 'TypeClass')



from typing import TypeVar, Callable

from lenses import Lens, lens

from amino import Maybe, List, __, Boolean, _, L

A = TypeVar('A')

_add = lambda l: __.map(lens().add_lens(l).add_lens)


def path_lens_pred(a: A, sub: Callable[[A], List[A]], lsub,
                   f: Callable[[A], bool]):
    g = lambda a: Boolean(f(lsub(a))).maybe(lsub(lens()))
    return path_lens(a, sub, g)


def path_lens(a: A, sub: Callable[[A], List[A]],
              f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
    return _path_lens(a, sub, f).smap(lens(a).tuple_)


def path_lens_unbound_pre(a: A, sub: Callable[[A], List[A]],
                          f: Callable[[A], Maybe[Lens]], pre: Callable):
    return (
        _path_lens(pre(a), sub, f) /
        (_ / pre(lens()).add_lens) /
        __.cons(lens())
    ).smap(lens().tuple_)


def path_lens_unbound(a: A, sub: Callable[[A], List[A]],
                      f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
    return _path_lens(a, sub, f).smap(lens().tuple_)


def _path_lens(a: A, sub: Callable[[A], List[A]],
               f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
    def go_sub():
        l, s = sub(lens()), sub(a)
        g = lambda b: _path_lens(b, sub, f)
        return _path_lens_list(s, g) / _add(l).cons(lens())
    return (f(a) / L(List)(lens(), _)).or_else(go_sub)


def _path_lens_list(fa: List[A], f: Callable[[A], Maybe[Lens]]) -> Maybe[Lens]:
    check = lambda a: f(a[1]) / (lambda b: (a[0], b))
    cat = lambda i, l: _add(lens()[i])(l)
    return fa.with_index.find_map(check).map2(cat)

__all__ = ('path_lens',)



import numbers

from amino import may


@may
def try_convert_int(data):
    if (
            isinstance(data, numbers.Number) or
            (isinstance(data, str) and data.isdigit())
    ):
        return int(data)

__all__ = ('try_convert_int',)
